#include "CANbus.hh"
#include "CANdefs.hh"

#define PUCK_IDLE 0 

CANbus::CANbus(){
  trq = NULL;
  pos = NULL;
  pucks = NULL;
  npucks = 0;
  pthread_mutex_init(&busmutex, NULL);
  pthread_mutex_init(&trqmutex, NULL);
  pthread_mutex_init(&posmutex, NULL);
  pthread_mutex_init(&runmutex, NULL);
}

int CANbus::start(){
  pthread_mutex_lock(&runmutex);
  bus_run = true;
  pthread_mutex_unlock(&runmutex);

  if(pthread_create(&canthread, NULL, canbus_handler, this) != 0){
    cerr << "CANbus::start: pthread_create failed." << endl;
    pthread_mutex_lock(&runmutex);
    bus_run = false;
    pthread_mutex_unlock(&runmutex);
    return OW_FAILURE;
  }
  return OW_SUCCESS;
}

int CANbus::stop(){
  pthread_mutex_lock(&runmutex);
  bus_run=false;
  pthread_mutex_unlock(&runmutex);

  if(pthread_join(canthread, NULL) != 0){
    cerr << "CANbus::stop: pthread_join failed." << endl;
    return OW_FAILURE;
  }
  return OW_SUCCESS;
}

int CANbus::init(){
#ifdef WRIST
  if(load("bus7.dat") == OW_FAILURE){
#else
  if(load("bus4.dat") == OW_FAILURE){
#endif
    cerr << "CANbus::init: load failed." << endl;
    return OW_FAILURE;
  }
   
  if(open() == OW_FAILURE){
    cerr << "CANbus::init: open failed." << endl;
    return OW_FAILURE;
  }

  if(check() == OW_FAILURE){
    cerr << "CANbus::init: check failed." << endl;
    return OW_FAILURE;
  }
  return OW_SUCCESS;
}

int CANbus::load(char* fn){
  ifstream is;
   
  is.open(fn);
  if(!is){
    cerr << "CANbus::load: couln't open file " << fn << endl;
    return OW_FAILURE;
  }
   
  is >> id;
  is >> npucks;
  pucks = new Puck[npucks+1];
  trq = new long[npucks+1];
  pos = new double[npucks+1];
  for(int p=1; p<=npucks; p++){
    pos[p] = 0.0;
    trq[p] = 0;
  }
   
  for(int p=1; p<=npucks; p++)
    is >> pucks[p];
  is.close();

  for(int p=1; p<=npucks; p++){
    if(groups[ pucks[p].group() ].insert(&pucks[p]) == OW_FAILURE){
      cerr << "CANbus::load: insert failed." << endl;
      return OW_FAILURE;
    }
  }
  return OW_SUCCESS;
}

int CANbus::open(){
  pthread_mutex_lock(&busmutex);
  if(canOpen(id, 0, TX_QUEUE_SIZE, RX_QUEUE_SIZE, 
	     TX_TIMEOUT, RX_TIMEOUT, &handle) != NTCAN_SUCCESS){
    cerr << "CANbus::open: canOpen failed." << endl;
    pthread_mutex_unlock(&busmutex);
    return OW_FAILURE;
  }
   
  // 1 = 1Mbps, 2 = 500kbps, 3 = 250kbps
  if(canSetBaudrate(handle, 0) != NTCAN_SUCCESS){
    cerr << "CANbus::opent: canSetBaudrate failed." << endl;
    pthread_mutex_unlock(&busmutex);
    return OW_FAILURE;
  }
  pthread_mutex_unlock(&busmutex);
    
  // Mask 3E0: 0000 0011 1110 0000
  // Messages sent directly to host
  if(allow_message(0x0000, 0x03E0) == OW_FAILURE){
    cerr << "CANbus::open: allow_message failed." << endl;
    return OW_FAILURE;
  }
  // Group 3 messages
  if(allow_message(0x0403, 0x03E0) == OW_FAILURE){
    cerr << "CANbus::open: allow_message failed." << endl;
    return OW_FAILURE;
  }
  // Group 6 messages
  if(allow_message(0x0406, 0x03E0) == OW_FAILURE){
    cerr << "CANbus::open: allow_message failed." << endl;
    return OW_FAILURE;
  }
   
  return OW_SUCCESS;
}

int CANbus::check(){
  long nodes[NUM_NODES+1];//, value;
  int online_pucks, reset_pucks, running_pucks;

  //Get status of pucks on bus
  usleep(10000);
  if(status(nodes) == OW_FAILURE){
    cerr << "CANbus::check: get_status failed." << endl;
    return OW_FAILURE;
  }
   
  cout<<endl;
  //count number of live pucks
  online_pucks = 0;
  for(int n=NODE_MIN; n<=NODE_MAX; n++){

    if(nodes[n] != STATUS_OFFLINE){      // display online nodes
      cout << "Node (id, status): (" << n << ", ";
      if(nodes[n] == STATUS_RESET){cout << "reset)" << endl;}
      else{cout << "running)" << endl;}
    }
	 
    if(nodes[n]!=STATUS_OFFLINE && n!=SAFETY_MODULE) 
      online_pucks++;
  }

  if(online_pucks==0){
    cerr << "The wam appears to be turned off." << endl;
    return OW_FAILURE;
  }
  if(online_pucks != npucks){
    cerr <<"Bus has " <<online_pucks <<" pucks. We expected " <<npucks<<endl;
    return OW_FAILURE;
  }  
  if(nodes[SAFETY_MODULE] == STATUS_OFFLINE){
    cerr << "CANbus::check: safety module is offline." << endl;
    return OW_FAILURE;
  }

  reset_pucks = 0;
  running_pucks = 0;
  for(int p=1; p<=npucks; p++){

    if(nodes[ pucks[p].id() ] == STATUS_RESET)
      reset_pucks++;
    else if(nodes[ pucks[p].id() ] != STATUS_OFFLINE)
      running_pucks++;
  }
   
  cout << endl << "Expected: " << npucks << "; Online: " << online_pucks
       << "; Running: " << running_pucks << "; Reset: " << reset_pucks<<endl;
      

  cout<<endl<<"**************************************"<<endl<<endl;

  if(running_pucks!=npucks && reset_pucks!=npucks){
    cerr << "Some of the pucks must have reset unexpectedly." <<endl;
    cerr << "ALL pucks should be in either running or reset state." <<endl;
    cerr << "Push the RESET button to fix that." <<endl<<endl;
    return OW_FAILURE;
  }
      
  for(int p=1; p<=npucks; p++){
    cout << "Checking puck " << pucks[p].id();
	 
    if(nodes[ pucks[p].id() ] == STATUS_RESET){
      cout << " (reset)" <<endl;
      cout << "\t->Waking up the puck...";
      if(wake_puck(p) == OW_FAILURE){
	cerr << "CANbus::check: wake_puck failed." << endl;
	return OW_FAILURE;
      }
      cout << "OK" << endl;
      usleep(100000);

      cout << "\t->Setting a PID thingy...";
      if(set_property(pucks[p].id(),PIDX, pucks[p].order(),true)==OW_FAILURE){
	cerr << "CANbus::check: set_prop failed (PID)." << endl;
	return OW_FAILURE;
      }
      cout << "OK" << endl;

      cout << "\t->Setting max torque...";
      if(set_property(pucks[p].id(), MT, 4731, true) == OW_FAILURE){
	cerr << "CANbus::check: set_prop failed (max torque)." << endl;
	return OW_FAILURE;
      }
      cout << "OK" << endl;
    }
    else{
      cout << " (running)" << endl;
      cout << "\t Setting idle mode...";
      // this was set to CONTROLLER_IDLE originally
      if(set_property(pucks[p].id(), MODE, PUCK_IDLE, true) == OW_FAILURE){
	cerr << "CANbus::check: set_prop failed (MODE)." << endl;
	return OW_FAILURE;
      }
      cout << "OK" << endl;
    }
  }
  cout<<endl<<"**************************************"<<endl;

  return OW_SUCCESS;
}

void CANbus::dump(){
  cout << endl << "Printing bus information." << endl;
  for(int g=GROUP_ID_MIN; g<=GROUP_ID_MAX; g++){
    if(groups[g].id() != GROUP_INVALID)
      cout<<groups[g]<<endl;
  }
}
   
int CANbus::allow_message(int id, int mask){
  int i;
  pthread_mutex_lock(&busmutex);
  for(i=0; i<2048; i++){
    if((i & ~mask) == id){
      if(canIdAdd(handle, i) != NTCAN_SUCCESS){
	cerr << "CANbus::allow_message: canIdAdd failed." << endl;
	pthread_mutex_unlock(&busmutex);
	return OW_FAILURE;
      }
    }
  }
  pthread_mutex_unlock(&busmutex);
  return OW_SUCCESS;
}

int CANbus::wake_puck(int p){
  if(set_property(pucks[p].id(), STAT, STATUS_READY, false) == OW_FAILURE){
    cerr << "CANbus::wake_puck: set_prop failed." << endl;
    return OW_FAILURE;
  }
  usleep(500000);                    // Wait 10ms for puck to initialize    
  return OW_SUCCESS;
}

// Got to make sure that nodes is NUM_NODES+1 long !!!
int CANbus::status(long* nodes){
  unsigned char msg[8];
  int msgid, msglen, nodeid, property;

  cout << endl << "Probing for nodes on the bus..." <<endl<<endl;

  for(int n=NODE_MIN; n<=NODE_MAX; n++){
    msg[0] = (unsigned char)STAT;   // Compile the packet
    nodes[n] = STATUS_OFFLINE;      // Initialize node as offline

    pthread_mutex_lock(&busmutex);
    if(send(NODE2ADDR(n), msg, 1, true) == OW_FAILURE){
      cerr << "CANbus::status: send failed." << endl;
      pthread_mutex_unlock(&busmutex);
      return OW_FAILURE;
    }
    usleep(1000);
    if(read(&msgid, msg, &msglen, false) == OW_FAILURE){
      cout<<"CANbus::status: node "<<NODE2ADDR(n)<<" didn't anwser."<<endl;
    }
    else{
      if(parse(msgid, msg, msglen,&nodeid,&property,&nodes[n])==OW_FAILURE){
	cerr << "CANbuss::status: parse failed." << endl;
	pthread_mutex_unlock(&busmutex);
	return OW_FAILURE;
      }
    }
    pthread_mutex_unlock(&busmutex);

  }
  cout<<endl<<"**************************************"<<endl;

  return OW_SUCCESS;
}

int CANbus::set_property(int32_t nid, int32_t property, long value,bool check){
  uint8_t msg[8];
  long response;
  int32_t msglen;
   
  if(compile(property, value, msg, &msglen) ==OW_FAILURE){
    cerr << "CANbus::set_property: compile failed." << endl;
    return OW_FAILURE;
  }
  msg[0] |= (uint8_t)0x80; // Set the 'Set' bit

  pthread_mutex_lock(&busmutex);
  if(send(NODE2ADDR(nid), msg, msglen, true) == OW_FAILURE){
    cerr << "CANbus::set_property: send failed." << endl;
    pthread_mutex_unlock(&busmutex);
    return OW_FAILURE;
  }
  pthread_mutex_unlock(&busmutex);
   
  if(check){
    // Get the new value of the property
    if(get_property(nid, property, &response) == OW_FAILURE){
      cerr << "CANbus::set_property: get_prop failed." << endl;
      return OW_FAILURE;
    }

    // Compare response to value
    if(response != value){
      cerr << "CANbus::set_property: response and value mismatch." << endl;
      return OW_FAILURE;
    }
  }   

  return OW_SUCCESS;
}

int CANbus::get_property(int32_t nid, int32_t property, long* value){
  uint8_t msg[8];
  int32_t msgid, msglen, nodeid, prop;

  msg[0] = (uint8_t)property;   

  pthread_mutex_lock(&busmutex);
  if(send(NODE2ADDR(nid), msg, 1, true) == OW_FAILURE){
    cerr << "CANB::get_prop: send failed." << endl;
    pthread_mutex_unlock(&busmutex);
    return OW_FAILURE;
  }
  if(read(&msgid, msg, &msglen, true) == OW_FAILURE){
    cerr << "CANbus::get_prop: read failed." << endl;
    pthread_mutex_unlock(&busmutex);
    return OW_FAILURE;
  }
  pthread_mutex_unlock(&busmutex);
   
  // Parse the reply
  if(parse(msgid, msg, msglen, &nodeid, &prop, value) == OW_FAILURE){
    cerr << "CANbus::get_prop: parse failed." << endl;
    return OW_FAILURE;
  }
      
  // Check that the ids and properties match
  if(nodeid!=nid || prop!=property){
    cerr << "CANbus::get_prop: id or property mismatch." << endl;
    return OW_FAILURE;
  }   
  return OW_SUCCESS;
}

int CANbus::send_torques(long* torques){
  pthread_mutex_trylock(&trqmutex);
  for(int p=1; p<=npucks; p++)
    trq[p] = torques[p];
  pthread_mutex_unlock(&trqmutex);

#ifndef __CANSERVER__
  if(send_torques() == OW_FAILURE){
    cerr << "CANbus::send_torques: send_torques failed." << endl;
    return OW_FAILURE;
  }
#endif

  return OW_SUCCESS;
}

int CANbus::send_torques(){
  uint8_t msg[8];
  int torques[NUM_ORDERS+1];
  Puck* puck;
  
  pthread_mutex_lock(&trqmutex);
  for(int g=GROUP_ID_MIN; g<=GROUP_ID_MAX; g++){
    if(groups[g].id() != GROUP_INVALID){

      for(int p=PUCK_ORDER_MIN; p<=PUCK_ORDER_MAX; p++){
	    
	puck = groups[g].puck(p); 
	if(puck){ 
	  torques[p] = clip(trq[puck->motor()], 
			    Puck::MIN_TRQ[puck->id()], 
			    Puck::MAX_TRQ[puck->id()] );
	}
	else{
	  torques[p] = 0;	    
	}
      }
      msg[0] = TORQ | 0x80; 
      msg[1] = (uint8_t)(( torques[1]>>6)&0x00FF);
      msg[2] = (uint8_t)(((torques[1]<<2)&0x00FC) | ((torques[2]>>12)&0x0003));
      msg[3] = (uint8_t)(( torques[2]>>4)&0x00FF);
      msg[4] = (uint8_t)(((torques[2]<<4)&0x00F0) | ((torques[3]>>10)&0x000F));
      msg[5] = (uint8_t)(( torques[3]>>2)&0x00FF);
      msg[6] = (uint8_t)(((torques[3]<<6)&0x00C0) | ((torques[4]>>8) &0x003F));
      msg[7] = (uint8_t)(  torques[4]    &0x00FF);
      
      pthread_mutex_lock(&busmutex);
      if(send(GROUPID(groups[g].id()), msg, 8, true) == OW_FAILURE){
	cerr << "CANbus::set_torques: send failed." << endl;
	pthread_mutex_unlock(&busmutex);
	return OW_FAILURE;
      }
      pthread_mutex_unlock(&busmutex);
    }
  }
  pthread_mutex_unlock(&trqmutex);

  return OW_SUCCESS;
}

int CANbus::read_torques(long* mtrq){
  uint8_t  msg[8];
  long int value;
  int32_t msgid, msglen, nodeid, property;

  // Compile the packet
  msg[0] = (uint8_t)TORQ;

  pthread_mutex_lock(&busmutex);
  if(send(GROUPID(0), msg, 1, true) == OW_FAILURE){
    cerr << "CANbus::get_positions: send failed." << endl;;
    pthread_mutex_unlock(&busmutex);
    return OW_FAILURE;
  }
  for(int p=1; p<=npucks; ){
    if(read(&msgid, msg, &msglen, true) == OW_FAILURE){
      cerr << "CANbus::get_positions: read failed." << endl;
      pthread_mutex_unlock(&busmutex);
      return OW_FAILURE;
    }
    if(parse(msgid, msg, msglen, &nodeid, &property, &value) == OW_FAILURE){
      cerr << "CANbus::get_positions: parse failed." << endl;
      pthread_mutex_unlock(&busmutex);
      return OW_FAILURE;
    }
    if(property == TORQ){
      trq[nodeid] = value;
      p++;
    }
  }
  pthread_mutex_unlock(&busmutex);

  return OW_SUCCESS;
}

int CANbus::read_positions(double* positions){
#ifndef __CANSERVER__
  if(read_positions() == OW_FAILURE){
    cerr << "CANbus::read_positions: read_positions failed." << endl;
    return OW_FAILURE;
  }
#endif

  for(int p=1; p<=npucks; p++)
    positions[p] = pos[p];

  return OW_SUCCESS;
}

int CANbus::read_positions(){
  uint8_t  msg[8];
  long data[NUM_NODES+1], value;
  int32_t msgid, msglen, property, nodeid;

  // Compile the packet
  msg[0] = (uint8_t)AP;

  pthread_mutex_lock(&busmutex);

  if(send(GROUPID(0), msg, 1, true) == OW_FAILURE){
    cerr << "CANbus::get_positions: send failed." << endl;;
    pthread_mutex_unlock(&busmutex);
    return OW_FAILURE;
  }

  pthread_mutex_lock(&posmutex);

  for(int p=1; p<=npucks; ){

    if(read(&msgid, msg, &msglen, true) == OW_FAILURE){
      cerr << "CANbus::get_positions: read failed." << endl;
      pthread_mutex_unlock(&posmutex);
      pthread_mutex_unlock(&busmutex);
      return OW_FAILURE;
    }

    if(parse(msgid, msg, msglen, &nodeid, &property, &value) == OW_FAILURE){
      cerr << "CANbus::get_positions: parse failed." << endl;
      pthread_mutex_unlock(&posmutex);
      pthread_mutex_unlock(&busmutex);
      return OW_FAILURE;
    }
    if(property == AP){
      data[nodeid] = value;
      p++;
    }
  }

  pthread_mutex_unlock(&posmutex);
  pthread_mutex_unlock(&busmutex);

  // convert the results
  for(int p=1; p<=npucks; p++)
    pos[ pucks[p].motor() ] = 2.0*M_PI*( (double) data[ pucks[p].id() ] )/ 
                                       ( (double) pucks[p].CPR() );

  return OW_SUCCESS;
}

int CANbus::send_positions(double* mpos){
  long position;

  //
  // WARNING
  // If the next call crashes we should put back the SAFETY_MODULE back on
  //
  if(set_property(SAFETY_MODULE, IFAULT, 8, true) == OW_FAILURE){
    cerr << "CANbus::send_positions: set_property(1) failed." << endl;
    return OW_FAILURE;
  }

  for(int p=1; p<=npucks; p++){
    position = (long)floor( mpos[ pucks[p].motor() ]*
			    pucks[p].CPR()/(2.0*M_PI) );

    if(set_property(pucks[p].id(), AP, position, true) == OW_FAILURE){
      cerr << "CANbus::send_positions: set_property failed." << endl;
      return OW_FAILURE;
    }
  }
   
  if(set_property(SAFETY_MODULE, ZERO, 1, true) == OW_FAILURE){
    cerr << "CANbus::send_positions: set_property(2) failed." << endl;
    return OW_FAILURE;
  }

  return OW_SUCCESS;
}

int CANbus::parse(int32_t msgid, uint8_t* msg, int32_t msglen,
		  int32_t* nodeid, int32_t* property, long* value){
  int i;
  int dataHeader;

  *nodeid = ADDR2NODE(msgid);
  if(*nodeid == -1){
    cerr<<"CANbus::parse: invalid node id."<<endl;
    return OW_FAILURE;
  }
   
  dataHeader = ((msg[0] >> 6) & 0x0002) | ((msgid & 0x041F) == 0x0403);

  switch (dataHeader){
      
  case 3:  // Data is a packed 22-bit position, SET 
    *value = 0x00000000;
    *value |= ( (long)msg[0] << 16) & 0x003F0000;
    *value |= ( (long)msg[1] << 8 ) & 0x0000FF00;
    *value |= ( (long)msg[2] )      & 0x000000FF;
      
    if(*value & 0x00200000) // If negative 
      *value |= 0xFFC00000; // Sign-extend 
      
    *property = AP;
    break;

  case 2:  // Data is normal, SET 
    *property = msg[0] & 0x7F;

    //
    // Store the value
    // second byte of message is zero (for DSP word alignment) 
    //
    *value = 0;
    for(i=0; i<msglen-2; i++)
      *value |= ((unsigned long)msg[i + 2] << (i * 8))
	& (0x000000FF << (i * 8));
      
    if (*value & (1 << ((i*8) - 1)))
      *value |= 0xFFFFFFFF << (i * 8); // Sign extend the value 
    break;

  case 0:  // Assume firmware request (GET) 
      
    // A negative (or zero) property means GET
    *property = -(msg[0] & 0x7F); 
    *value = 0;
    break;
      
  default:
    cerr<<"CANbus::parse: Illegal message header."<<endl;
    return OW_FAILURE;
  }
  return OW_SUCCESS;
}

int CANbus::compile(int32_t property, long value, 
		    uint8_t *msg, int32_t *msglen){
  int i;
   
  // Check the property
  if(PROP_END < property){
    cerr<<"CANbus::compile: invalid property."<<endl;
    return OW_FAILURE;
  }
   
  msg[0] = (uint8_t)property; // Insert the property
  msg[1] = 0;                 // To align the values for the tater's DSP
   
  // Append the value 
  for(i=2; i<6; i++){
    msg[i] = (uint8_t)(value & 0x000000FF);
    value >>= 8;
  }
   
  // Record the proper data length 
  *msglen = (dataType[property] & 0x0007) + 2;
   
  return OW_SUCCESS;
}

int CANbus::read(int32_t* msgid, uint8_t* msg, int32_t* msglen, bool block){
  CMSG cmsg;
  int32_t len = 1;
  int i, err;
   
  //rt_make_soft_real_time();
  if(block){
    if((err=canRead(handle, &cmsg, &len, NULL)) != NTCAN_SUCCESS){
      cerr << "CANbus::read: canRead failed: " << hex << err << endl;
      return OW_FAILURE;
    }
  }
  else{
    if((err=canTake(handle, &cmsg, &len)) != NTCAN_SUCCESS){
      cerr << "CANbus::read: canTake failed: " << hex << err << endl;
      return OW_FAILURE;
    }
  }
  //rt_make_hard_real_time();

  // Just to avoid printing anoying output during status checkup
  if(len == 0){
    return OW_FAILURE;
  }
  else if(len != 1){
    cerr<<"CANbus::read: received a message of length: "<<len<<endl;
    return OW_FAILURE;
  }

  *msgid = cmsg.id;
  *msglen = cmsg.len;
  for(i=0; i<*msglen; i++)
    msg[i] = cmsg.data[i];
   
  return OW_SUCCESS;
}

int CANbus::send(int32_t msgid, uint8_t* msg, int32_t msglen, bool block){
  CMSG cmsg;
  int32_t len = 1;
  int i, err;

  cmsg.id = msgid;
  cmsg.len = (uint8_t)(msglen & 0x0F);
  for(i=0; i<msglen; i++)
    cmsg.data[i] = msg[i];

  if(block == true){
    if( (err=canWrite(handle, &cmsg, &len, NULL)) != NTCAN_SUCCESS){
      cerr << "CANbus::send: canWrite failed: " << hex << err << endl;
      return OW_FAILURE;
    }
  }
  else{
    if( (err=canSend(handle, &cmsg, &len)) != NTCAN_SUCCESS){
      cerr << "CANbus::send: canSend failed: " << hex << err << endl;
      return OW_FAILURE;
    }
  }
  //rt_make_hard_real_time();
  return OW_SUCCESS;
}

// set safety limits (does it really work?)
// it always sets the same property!!
int CANbus::limits(double jointVel, double tipVel, double elbowVel){
  long conversion;
   
  if(0<jointVel && jointVel<7){           // If the vel (rad/s) is reasonable
    conversion = (long)(jointVel*0x1000); // Convert to Q4.12 rad/s
    if(set_property(SAFETY_MODULE, VL2, conversion, true) == OW_FAILURE){
      cerr << "WAM::set_limits: set_prop failed." << endl;
      return OW_FAILURE;
    }
  }
   
  if(0<tipVel && tipVel<7){               // If the vel (m/s) is reasonable
    conversion = (long)(tipVel*0x1000);   // Convert to Q4.12 rad/s
    if(set_property(SAFETY_MODULE, VL2, conversion, true) == OW_FAILURE){
      cerr << "WAM::set_limits: set_prop failed." << endl;
      return OW_FAILURE;
    }
  }
   
  if(0<elbowVel && elbowVel<7){           // If the vel (m/s) is reasonable
    conversion = (long)(elbowVel*0x1000); // Convert to Q4.12 rad/s
    if(set_property(SAFETY_MODULE, VL2, conversion, true) == OW_FAILURE){
      cerr << "WAM::set_limits: set_prop failed." << endl;
      return OW_FAILURE;
    }
  }
  return OW_SUCCESS;
}

int CANbus::run(){
  int r;
  pthread_mutex_lock(&runmutex);
  r = bus_run;
  pthread_mutex_unlock(&runmutex);
  return r;
}


void CANbus::printpos(){
  for(int p=1; p<=npucks; p++)
    cout << pos[p] << " ";
}

void* canbus_handler(void* argv){
  CANbus* bus;
  RT_TASK* task;

#ifdef __LATENCY__
#define ARRAY_SIZE 5000
  RTIME t;
  long long array[ARRAY_SIZE];
  int i=0;
  FILE* fp;
#endif

  bus = (CANbus*)argv;

  task=rt_task_init_schmod(nam2num("CANBUS"), 1, 0, 0, SCHED_FIFO, 1);
  if(task == NULL){
    cerr << "canbus_handler: rt_task_init_schmod failed." << endl;
    return NULL;
  }
  if(mlockall(MCL_CURRENT | MCL_FUTURE) == -1){
    perror("canbus_handler: mlockall failed: ");
    return NULL;
  }
  rt_make_hard_real_time();

  while(bus->run()){

    if(bus->read_positions() == OW_FAILURE){
      cerr << "canbus_handler: read_positions failed." << endl;
      rt_task_delete(task);
      return NULL;
    }
    if(bus->send_torques() == OW_FAILURE){
      cerr << "canbus_handler: send_torques failed." << endl;
      rt_task_delete(task);
      return NULL;
    }
    
#ifdef __LATENCY__
    t = rt_get_cpu_time_ns();
    if(i < ARRAY_SIZE)
      array[i] = t;
    if(i == ARRAY_SIZE)
      cout << "CANbus latency recording done." << endl;
    i++;
#endif

  }

#ifdef __LATENCY__
  fp = fopen("canlatencies", "w");
  for(i=0; i<ARRAY_SIZE; i++)
    fprintf(fp, "%lld ", array[i]);
  fclose(fp);
#endif

  rt_task_delete(task);

  return argv;
}


