#include <sys/time.h>
#include "Dynamics.hh"

static double g = 9.81;
static R3 z0(0,0,1);

/*
 * Recursive Newton-Euler method (Luh, Walker, Paul)
 */
void RNE(double tau[Link::Ln+1], 
	 Link links[Link::Ln+1], 
	 double qd[Link::Ln+1], 
	 double qdd[Link::Ln+1], R6& fext){

  R3 w, wd, v, vd, vdhat;
  R3 n, f, N[Link::Ln+1], F[Link::Ln+1];

  vd = g*z0;

  for(int i=Link::L1; i<=Link::Ln; i++){

    SO3 A      = !((SO3)links[i]);
    Inertia I  =        links[i].I();
    R3 ps      =        links[i].ps();
    R3 s       =        links[i].s();
    double m   =        links[i].m();
    
    wd = A*( wd + (z0*qdd[i]) + (w^(z0*qd[i])) );
    w  = A*( w  + (z0*qd [i]) );
    vd = (wd^ps) + (w^(w^ps)) + A*vd;

    vdhat = (wd^s) + (w^(w^s)) + vd;
    F[i] = m*vdhat;
    N[i] = (I*wd) + (w^(I*w));
  }

  /*
  f[R3::X] = fext[Wrench::FX]; 
  f[R3::Y] = fext[Wrench::FY];
  f[R3::Z] = fext[Wrench::FZ];
  n[R3::X] = fext[Wrench::MX];
  n[R3::Y] = fext[Wrench::MY];
  n[R3::Z] = fext[Wrench::MZ];
  */

  for(int i=Link::Ln; Link::L1<=i; i--){
    SO3 A;
    R3 ps = links[i].ps();
    R3 s  = links[i].s();

    if(i != Link::Ln)
      A = (SO3)links[i+1];

    //n = A*( n + (((!A)*ps)^f) ) + ((ps+s)^F[i]) + N[i];
    //f = A*f + F[i];

    f = A*f + F[i];
    n = A*n + (ps^f) + (s^F[i]) + N[i];  // ask me for this simplification
    A = !((SO3)links[i]);
    tau[i] = n*(A*z0);
  }
}

/*
 * Use to get the Coriolis/Centrifugal + gravity load
 * Based on RNE with all acceleration equal to 0 (Walker, Orin)
 */
void CCG(double ccg[Link::Ln+1], Link links[Link::Ln+1],double qd[Link::Ln+1]){
  R3 w, wd, v, vd, vdhat;
  R3 n, f, N[Link::Ln+1], F[Link::Ln+1];
  
  vd = g*z0;

  for(int i=Link::L1; i<=Link::Ln; i++){

    SO3 A      = !((SO3)links[i]);
    Inertia I  =        links[i].I();
    R3 ps      =        links[i].ps();
    R3 s       =        links[i].s();
    double m   =        links[i].m();

    wd = A*( wd + (w^(z0*qd[i])) );
    w  = A*( w  + (z0*qd [i]) );
    vd = (wd^ps) + (w^(w^ps)) + A*vd;

    vdhat = (wd^s) + (w^(w^s)) + vd;
    F[i] = m*vdhat;
    N[i] = (I*wd) + (w^(I*w));
  }

  for(int i=Link::Ln; Link::L1<=i; i--){
    SO3 A;
    R3 ps = links[i].ps();
    R3 s  = links[i].s();

    if(i != Link::Ln)
      A = (SO3)links[i+1];

    f = A*f + F[i];
    n = A*n + (ps^f) + (s^F[i]) + N[i];
    A = !((SO3)links[i]);
    ccg[i] = n*(A*z0);
  }
}

/*
 * Use to compute the wd and vd. Would be cleaner if it returned a twist...
 */
R6 acceleration( Link links[Link::Ln+1], 
		 double qd[Link::Ln+1], 
		 double qdd[Link::Ln+1] ){
  R3 w, wd, vd;

  for(int i=Link::L1; i<=Link::Ln; i++){

    SO3 A      = !((SO3)links[i]);
    R3 ps      =        links[i].ps();
    
    wd = A*( wd + (z0*qdd[i]) + (w^(z0*qd[i])) );
    w  = A*( w  + (z0*qd [i]) );
    vd = (wd^ps) + (w^(w^ps)) + A*vd;
  }

  return R6(vd, wd);
}

/*
 * Computes the bias acceleration Jd*qd (as in [vdd;wdd] = Jd*qd + J*qdd
 */
R6 bias_acceleration(Link links[Link::Ln+1], double qd[Link::Ln+1]){
  R3 w, wd, vd;

  for(int i=Link::L1; i<=Link::Ln; i++){

    SO3 A      = !((SO3)links[i]);
    R3 ps      =        links[i].ps();
    
    wd = A*( wd + ( w^(z0*qd[i]) ) );
    w  = A*( w  + (    z0*qd[i]  ) );
    vd = (wd^ps) + (w^(w^ps)) + A*vd;
  }

  return R6(vd, wd);
}

/*
 * Computation of the joint space inertia matrix (Walker, Orin)
 * Because the it is symmetric, only the top half is filled.
 * So in fortran (column major) that becomes the lower half
 * Takes about 10usec
 */
void JSinertia(double A[Link::Ln][Link::Ln], Link links[Link::Ln+1]){
  Inertia I(1.0, 0.0, 0.0, 1.0, 0.0, 1.0);
  int j = Link::Ln;
  SO3 iAj = (SO3)links[j];
  R3 jsj =       links[j].s();
  R3 jpsj =      links[j].ps();
  double mj =    links[j].m();
  double Mj = mj;

  Inertia jEj = links[j].I();
  R3 icj = iAj*(jsj + jpsj);

  Inertia iEj = iAj*jEj*(!iAj);

  R3 iFj = z0^(Mj*icj);
  R3 iNj = iEj*z0;
  
  R3 ifj = iFj;
  R3 inj = iNj + (icj^iFj);
  
  A[j-1][j-1] = inj[R3::Z];

  for(int i=j-1; Link::L1<=i; i--){
    iAj     = (SO3)links[i];
    R3 ipsi =      links[i].ps();

    inj = iAj * (inj + (ipsi^ifj));  // order is important
    ifj = iAj * ifj;

    A[i-1][j-1] = inj[R3::Z];    // the -1 is to ajust the indexes
  }

  for(j=Link::Ln-1; Link::L1<=j; j--){
    double Mk = Mj;
    R3 jck = icj;
    Inertia jEk = iEj;

    iAj         = (SO3)links[j];
    jsj         =      links[j].s();
    Inertia jJj =      links[j].I();
    jpsj        =      links[j].ps();

    mj = links[j].m();
    Mj = Mj + mj;

    R3 jcj = (1.0/Mj) * ( mj*(jsj+jpsj) + Mk*(jck+jpsj) );
    icj = iAj * jcj;

    R3 v1(jck+jpsj-jcj);
    R3 v2(jsj+jpsj-jcj);
    Inertia J1(v1[R3::X]*v1[R3::X], v1[R3::X]*v1[R3::Y], v1[R3::X]*v1[R3::Z],
	       v1[R3::Y]*v1[R3::Y], v1[R3::Y]*v1[R3::Z], v1[R3::Z]*v1[R3::Z]);
    Inertia J2(v2[R3::X]*v2[R3::X], v2[R3::X]*v2[R3::Y], v2[R3::X]*v2[R3::Z],
	       v2[R3::Y]*v2[R3::Y], v2[R3::Y]*v2[R3::Z], v2[R3::Z]*v2[R3::Z]);

    iEj = iAj * ( jEk + Mk*( I*(v1*v1) - J1 ) + 
		  jJj + mj*( I*(v2*v2) - J2 ) ) * (!iAj);
    
    iFj = z0^(Mj*icj);
    iNj = iEj*z0;

    ifj = iFj;
    inj = iNj + (icj^iFj);
    A[j-1][j-1] = inj[R3::Z];

    for(int i=j-1; Link::L1<=i; i--){
      iAj     = (SO3)links[i];
      R3 ipsi =      links[i].ps();
     
      inj = iAj * (inj + (ipsi^ifj)); // order is important
      ifj = iAj * ifj;

      A[i-1][j-1] = inj[R3::Z];
    }
  }
}

/*
 * Computation of the worspace space inertia matrix (Khatib)
 * Only the upper half is computed (lower part in fortran)
 * Takes about 57usec
 */
void WSinertia(double Ac[6][6], Link links[Link::Ln+1]){
  char SIDE = 'R';
  char UPLO = 'L';
  char TRANST = 'T';  
  char TRANSN = 'N';
  double ALPHA = 1.0; 
  double BETA = 0.0;
  int NEQS = 6; 
  int NJOINTS = Joint::Jn;
  int INFO;

  double A[Joint::Jn][Joint::Jn];       // Joint space inertia matrix
  int LDA = Joint::Jn;
  JSinertia(A, links);

  double J [Joint::Jn][6];              // Column major Jacobian
  int LDJ = 6;
  JacobianNF(J, links);

  // A^-1
  // Factorize the symmetric positive definite matrix
  dpotrf(&UPLO, &NJOINTS, &A[0][0], &LDA, &INFO);
  if(INFO<0)
    cerr << "WSinertia::dpotrf1: The " << INFO << " argument is illegal.\n";
  else if(0<INFO)
    cerr << "WSinertia::dpotrf1: The matrix is not positive definite.\n";
  // invert
  dpotri(&UPLO, &NJOINTS, &A[0][0], &LDA, &INFO);
  if(INFO<0)
    cerr << "WSinertia::dpotri1: The " << INFO << " argument is illegal.\n";
  else if(0<INFO)
    cerr << "WSinertia::dpotri1: The matrix is singular.\n";

  // J*A^-1
  double JAi [Joint::Jn][6];
  int LDJAi = 6;
  dsymm(&SIDE, &UPLO, &NEQS, &NJOINTS, &ALPHA, 
	&A  [0][0], &LDA, 
	&J  [0][0], &LDJ, &BETA, 
	&JAi[0][0], &LDJAi);

  // J*A^-1*J'
  int LDAc = 6;
  dgemm(&TRANSN, &TRANST, &NEQS, &NEQS, &NJOINTS, &ALPHA, 
	&JAi[0][0], &LDJAi, 
	&J  [0][0], &LDJ, &BETA, 
	&Ac [0][0], &LDAc);

  // (J*A^-1*J')^-1
  // Factorize the symmetric matrix
  dpotrf(&UPLO, &NEQS, &Ac[0][0], &LDAc, &INFO);
  if(INFO<0)
    cerr << "WSinertia::dpotrf2: The " << INFO << " argument is illegal.\n";
  else if(0<INFO)
    cerr << "WSinertia::dpotrf2: The matrix is not positive definite.\n";
  // invert
  dpotri(&UPLO, &NEQS, &Ac[0][0], &LDAc, &INFO);
  if(INFO<0)
    cerr << "WSinertia::dpotri2: The " << INFO << " argument is illegal.\n";
  else if(0<INFO)
    cerr << "WSinertia::dpotri2: The matrix is singular.\n";
}

/*
 * Operation Space dynamics with MKL libraries
 * Takes about 67us. And that kind of suck. Because the control loop of the
 * WAM takes about 930usec just to receive joint positions and send torques,
 * which brings the total around just above 1ms with this control scheme.
 * Therefore, the control loop must run at 500Hz instead of 1kHz.
 * So, either spend time figuring out a better/faster driver/library that will
 * speed up the CAN stuff. Or figure out how to shave a few extra usecs from 
 * this (but I think it is already fairly optimized)
 * I mean, it's not that bad, most of the control iterations are about 990usec,
 * however too many of them go over 1ms which then screws up RTAI and pump
 * even more CPU.
 */

void WSdynamics(double trq[Link::Ln+1], 
		Link links[Link::Ln+1], 
		double qd[Link::Ln+1], R6& F){

  char UPLO = 'L';
  char TRANST = 'T';
  int NJOINTS = Joint::Jn;
  int NEQS = 6;
  int INC = 1;

  double ALPHA =  1.0;
  double BETA  =  0.0;
  double GAMMA = -1.0;
  
  for(int j=Joint::J1; j<=Joint::Jn; j++)
    trq[j] = 0.0;

  double ccg[Joint::Jn+1];           // coriolis/centrifugal + gravity
  CCG(ccg, links, qd);

  // This is little hack to avoid singularities in Ac when using the WAM
  // in interactive mode (no forces). Otherwise the WAM will jerk when q4=0
  if(0.0 < F.norm()){
    double Ac[6][6];                 // Cartesian inertia matrix (lower half)
    int LDAc = 6;
    double J[Joint::Jn][6];          // Column major Jacobian
    int LDJ = 6;

    WSinertia(Ac, links);
    JacobianNF(J, links);

    // Ac*F
    R6 AcF;
    dsymv(&UPLO, &NEQS, &ALPHA,      // symmetric matrix * vector
	  &Ac[0][0],          &LDAc,
	  (double*)F,   &INC, &BETA,
	  (double*)AcF,       &INC);
    
    // J'*Ac*F                       // matrix * vector
    dgemv(&TRANST, &NEQS,  &NJOINTS, &ALPHA,
	  &J[0][0],        &LDJ,
	  (double*)AcF,    &INC, &BETA,
	  &trq[Joint::J1], &INC);

    // Ac*h
    R6 h = bias_acceleration(links, qd);
    R6 Ach;                          // symmetric matrix * vector
    dsymv(&UPLO, &NEQS, &ALPHA,
	  &Ac[0][0],    &LDAc,
	  (double*)h,   &INC, &BETA,
	  (double*)Ach, &INC);
      
    // ccg - J'*Ac*h
    dgemv(&TRANST, &NEQS,  &NJOINTS, &GAMMA,
	  &J[0][0],        &LDJ,
	  (double*)Ach,    &INC, &ALPHA,
	  &ccg[Joint::J1], &INC);
  }

  for(int j=Joint::J1; j<=Joint::Jn; j++)    trq[j] += ccg[j];  
}

void JSdynamics(double trq[Link::Ln+1], 
		Link   links[Link::Ln+1], 
		double qd[Link::Ln+1], 
		double qdd[Link::Ln+1]){

  double ssq = 0.0;
  for(int j=Joint::J1; j<=Joint::Jn; j++){
    trq[j] = 0.0;
    ssq += qdd[j]*qdd[j];
  }

  double ccg[Joint::Jn+1];           // coriolis/centrifugal + gravity
  CCG(ccg, links, qd);

  if(0.0 < ssq){
    char UPLO = 'L';
    int N = Joint::Jn;
    int LDA = 7;
    int INC = 1;
    double ALPHA = 1.0;
    double BETA = 0.0;

    double A[Joint::Jn][Joint::Jn];       // Joint space inertia matrix
    JSinertia(A, links);

    dsymv(&UPLO,           &N,  &ALPHA,   // symmetric matrix * vector
	  &A[0][0],        &LDA,
	  &qdd[Joint::J1], &INC, &BETA,
	  &trq[Joint::J1], &INC);
  }

  for(int j=Joint::J1; j<=Joint::Jn; j++)    trq[j] += ccg[j];  
}
