#include "Trajectory.hh"
#include "Joint.hh"
#include <math.h>

#ifndef __JOINTSTRAJ_HH__
#define __JOINTSTRAJ_HH__

/*
 * Implementa a joint trajectory. This is pretty easy as it just evaluates
 * whichever profile you have. I guess you could directly use a profile instead
 * but, I like consistent code and since you wouldn't be able to do the same
 * for Cartesian trajectory, then this class is the equivalent of SE3Trajectory
 */

class JointsTraj : public Trajectory{
protected:

  Profile *prof[Joint::Jn+1];
  double q1[Joint::Jn+1];     // initial positions
  double u[Joint::Jn+1];      // directions of the movement

public:

  JointsTraj( const double Q1[Joint::Jn+1], 
	      const double Q2[Joint::Jn+1], 
	      Profile* p[Joint::Jn+1] ) : Trajectory(){

    for(int j=Joint::J1; j<=Joint::Jn; j++){
      q1[j] = Q1[j];
      if(Q1[j] < Q2[j]) u[j] = +1;  // moving in the positive direction
      else              u[j] = -1;  // moving in the negative direction
      prof[j] = p[j];
      prof[j]->init( 0.0, fabs(Q2[j]-Q1[j]) );
    }
    stop();
  }

  // word of caution! the trajectory DELETE the profile
  // That implies that 1) you must allocate a profile and 2) don't delete it
  ~JointsTraj(){
    for(int j=Joint::J1; j<=Joint::Jn; j++)
      delete prof[j];
  }

  void run (){
    lock();
    for(int j=Joint::J1; j<=Joint::Jn; j++)
      prof[j]->run();
    unlock();
  }

  void stop(){
    lock();
    for(int j=Joint::J1; j<=Joint::Jn; j++)
      prof[j]->stop();
    unlock();
  }

  int state(){
    bool run = false;
    lock();
    for(int j=Joint::J1; j<=Joint::Jn; j++)
      if(prof[j]->state() == Profile::RUN){run = true;}
    unlock();
    if(run) return Trajectory::RUN;
    return Trajectory::STOP;
  }

  void evaluate(double y[Joint::Jn+1], 
		double yd[Joint::Jn+1], 
		double ydd[Joint::Jn+1], double dt){
    if(state() == Trajectory::RUN){
      lock();
      for(int j=Joint::J1; j<=Joint::Jn; j++){
	prof[j]->evaluate(y[j], yd[j], ydd[j], dt);
	y[j] = u[j]*y[j] + q1[j];
	yd[j] = u[j]*yd[j];
	ydd[j] = y[j]*ydd[j];
      }
      unlock();
    }
  }
};

#endif
